"""jirelease
infinidat jira project command-line tool

Usage:
    jirelease summary [--since-date=SINCE]
    jirelease list {project}
    jirelease release {project} {version}
    jirelease merge {project} {version} <target-version>
    jirelease delay {project} {version} <delta>
    jirelease reschedule {project} {version} <date>
    jirelease create {project} <target-version> [<delta>] [<description>]
    jirelease move {project} {version} (before | after) <target-version>
    jirelease archive {project} <version-regex>
    jirelease unarchive {project} <version-regex>
    jirelease rename {project} {version} <name>
    jirelease describe {project} {version} <description>

Options:
    summary                              list a summary of today's releases
    list                                 list unarchives releases
    release                              mark version as released
    merge                                move issues to target version and delete the merged one
    delay                                move the release date
    reschedule                           set a new release date
    create                               create a new release
    move                                 move a release up or down
    archive                              mark version as archived
    unarchive                            mark version as unarchived
    rename                               change name
    describe                             change description
    --project=PROJECT                    project key {project_default}
    --release=RELEASE                    version string {version_default}
    --since-date=SINCE                   since when [default: today]
"""
from __future__ import print_function


def _get_arguments(argv, environ):
    from .__version__ import __version__
    from docopt import docopt
    from munch import Munch
    project_default = "[default: {}]".format(environ["JISSUE_PROJECT"]) if "JISSUE_PROJECT" in environ else ""
    version_default = "[default: {}]".format(environ["JISSUE_VERSION"]) if "JISSUE_VERSION" in environ else ""
    doc_with_defaults = __doc__.format(project_default=project_default, version_default=version_default,
                                       project="[--project=PROJECT]" if project_default else "--project=PROJECT",
                                       version="[--release=RELEASE]" if version_default else "--release=RELEASE")
    arguments = Munch(docopt(doc_with_defaults, argv=argv, help=True, version=__version__))
    if environ.get("JISSUE_PROJECT") and not arguments.get("--project"):
        arguments["--project"] = environ["JISSUE_PROJECT"]
    if environ.get("JISSUE_VERSION") and not arguments.get("--release"):
        arguments["--release"] = environ["JISSUE_VERSION"]
    return arguments


def pretty_print_project_versions_in_order(project_name):
    from .jira_adapter import get_project, from_jira_formatted_datetime
    from prettytable import PrettyTable
    project = get_project(project_name)
    table = PrettyTable(["Name", "Description", "Release Date"])
    table.align = 'l'

    for version in reversed(project.versions):
        if version.archived:
            continue
        table.add_row([version.name if version.released else version.name + ' **' if getattr(version, 'overdue', False) else version.name + ' *',
                       getattr(version, 'description', ''), getattr(version, 'releaseDate', '')])
    print((table.get_string()))


def release_version(project_name, project_version):
    from .jira_adapter import get_version, to_jira_formatted_date
    from json import loads
    from datetime import datetime
    version = get_version(project_name, project_version)
    if version.released:
        raise AssertionError("version already released")
    unresolved_issue_count = list(loads(version._session.get(version.self + '/unresolvedIssueCount').text).values())[-1]
    if unresolved_issue_count:
        raise AssertionError("version has {} unresovled issues".format(unresolved_issue_count))
    if not version.releaseDate:
        version.update(releaseDate=to_jira_formatted_date(datetime.today()))
    version.update(released=True)


def merge_releases(project_name, project_version, target_version):
    from .jira_adapter import get_version
    version = get_version(project_name, project_version)
    target_version = get_version(project_name, target_version)
    version.delete(target_version.id, target_version.id)


def parse_deltastring(string):
    from datetime import timedelta
    from argparse import ArgumentTypeError
    DELTA_KEYWORD_ARGUMENTS = dict(w="weeks", d="days")
    keyword_argument = DELTA_KEYWORD_ARGUMENTS.get(string[-1] if string else "", "seconds")
    stripped_string = string.strip(''.join(list(DELTA_KEYWORD_ARGUMENTS.keys())))
    try:
        return timedelta(**{keyword_argument: abs(int(stripped_string))})
    except (ValueError, TypeError) as msg:
        raise ValueError("Invalid delta string: {!r}".format(string))


def delay_release(project_name, project_version, delta):
    from .jira_adapter import get_version, to_jira_formatted_date, from_jira_formatted_date
    version = get_version(project_name, project_version)
    new_release_date = from_jira_formatted_date(version.releaseDate) + parse_deltastring(delta)
    version.update(releaseDate=to_jira_formatted_date(new_release_date))


def reschedule_release(project_name, project_version, release_date):
    from .jira_adapter import get_version, to_jira_formatted_date, from_jira_formatted_date
    version = get_version(project_name, project_version)
    version.update(releaseDate=release_date if release_date else None)


def create_new_release(project_name, target_version, delta, description):
    from .jira_adapter import clear_cache, get_jira, get_project, from_jira_formatted_date, to_jira_formatted_date
    from pkg_resources import parse_version
    project = get_project(project_name)
    sorted_versions = sorted(project.versions, key=lambda version: parse_version(version.name))
    previous_version = sorted_versions[0]
    for item in sorted_versions:
        if parse_version(item.name) > parse_version(target_version):
            break
        previous_version = item
    if delta and not hasattr(previous_version, 'releaseDate'):
        raise AssertionError("previous version {} has no release date".format(previous_version.name))
    if delta:
        release_date = to_jira_formatted_date(from_jira_formatted_date(previous_version.releaseDate) + parse_deltastring(delta))
    else:
        release_date = None
    get_jira().create_version(target_version, project, releaseDate=release_date, description=description)
    clear_cache(get_project)
    move_release(project_name, target_version, after=True, target_version=previous_version.name)


def move_release(project_name, project_version, after, target_version):
    from .jira_adapter import get_version, get_project, JIRAError
    from json import dumps
    project = get_project(project_name)
    version = get_version(project_name, project_version)
    target_version = get_version(project_name, target_version)
    if not after: # before
        target_version = project.versions[project.versions.index(target_version)-1]
    url = version.self + '/move'
    response = version._session.post(url, headers={'content-type': 'application/json'}, data=dumps(dict(after=target_version.self)))
    if response.status_code != 200:
        raise JIRAError(url=url, status_code=response.status_code, text=response.reason)


def set_archive(project_name, project_version_regex, archived):
    from .jira_adapter import get_project
    from re import match
    for version in get_project(project_name).versions:
        if match(project_version_regex, version.name):
            version.update(archived=archived)


def set_name(project_name, project_version, name):
    from .jira_adapter import get_version
    version = get_version(project_name, project_version)
    version.update(name=name)


def set_description(project_name, project_version, description):
    from .jira_adapter import get_version
    version = get_version(project_name, project_version)
    version.update(description=description)


def datetime_to_date(item):
    from datetime import date
    return date(item.year, item.month, item.day)


def summary(since):
    from .jira_adapter import iter_projects, from_jira_formatted_date, to_jira_formatted_date
    from datetime import date
    from prettytable import PrettyTable

    table = PrettyTable(["Project", "Version", "Description", "Release Date"])
    table.align = 'l'
    today = date.today()
    since_date = today if since == 'today' else datetime_to_date(from_jira_formatted_date(since))

    for project in iter_projects():
        for version in reversed(project.versions):
            if version.archived:
                continue
            release_date_string = getattr(version, 'releaseDate', '')
            if not release_date_string:
                continue
            release_date = datetime_to_date(from_jira_formatted_date(release_date_string))
            if (release_date-since_date).days<0:
                continue
            if (release_date-today).days>0:
                continue

            table.add_row([project.name,
                           version.name if version.released else version.name + ' **' if getattr(version, 'overdue', False) else version.name + ' *',
                           getattr(version, 'description', ''), getattr(version, 'releaseDate', '')])

    print((table.get_string()))


def do_work(arguments):
    project_name = arguments['--project']
    project_version = arguments.get('--release')
    if arguments['summary']:
        summary(arguments.get("--since-date"))
    elif arguments['list']:
        pretty_print_project_versions_in_order(project_name)
    elif arguments['release']:
        release_version(project_name, project_version)
    elif arguments['merge']:
        merge_releases(project_name, project_version, arguments['<target-version>'])
    elif arguments['delay']:
        delay_release(project_name, project_version, arguments['<delta>'])
    elif arguments['reschedule']:
        reschedule_release(project_name, project_version, arguments['<date>'])
    elif arguments['create']:
        create_new_release(project_name, arguments['<target-version>'], arguments['<delta>'], arguments['<description>'])
    elif arguments['move']:
        move_release(project_name, project_version, arguments.get('after'), arguments['<target-version>'])
    elif arguments['archive']:
        set_archive(project_name, arguments['<version-regex>'], True)
    elif arguments['unarchive']:
        set_archive(project_name, arguments['<version-regex>'], False)
    elif arguments['rename']:
        set_name(project_name, project_version, arguments['<name>'])
    elif arguments['describe']:
        set_description(project_name, project_version, arguments['<description>'])


def _jiject(argv, environ):
    from sys import stderr
    from copy import deepcopy
    from jira import JIRAError
    from infi.execute import ExecutionError
    from .actions import choose_action
    from docopt import DocoptExit
    try:
        arguments = _get_arguments(argv, dict(deepcopy(environ)))
        return do_work(arguments)
    except DocoptExit as e:
        print(e, file=stderr)
        return 1
    except AssertionError as e:
        print(e, file=stderr)
        return 1
    except SystemExit as e:
        print(e, file=stderr)
        return 0
    except JIRAError as e:
        print(e, file=stderr)
    except ExecutionError as e:
        print(e.result.get_stderr(), file=stderr)
    return 1


def main():
    from os import environ
    from sys import argv
    return _jiject(argv[1:], environ)



from infi.pyutils.lazy import cached_function
from .config import Configuration
import requests
try:
    from urlparse import urljoin
except:
    from urllib.parse import urljoin
from .credential_store import ConfluenceCredentialsStore
from requests.auth import HTTPBasicAuth


@cached_function
def get_auth():
    config = Configuration.from_file()
    fqdn = config.confluence_fqdn
    credential_store = ConfluenceCredentialsStore()
    credentials = credential_store.get_credentials(fqdn)
    return HTTPBasicAuth(credentials.get_username(), credentials.get_password())


@cached_function
def get_headers():
    return {'Accept': 'application/json'}


@cached_function
def _get_confluence_uri(path):
    config = Configuration.from_file()
    if config.confluence_fqdn is None:
        raise Exception("Confluence FQDN not set in configuration file. Run 'jirelnotes config set'")
    return urljoin('https://{fqdn}/rest/'.format(fqdn=config.confluence_fqdn), path)


@cached_function
def _get_confluence_global_response(global_label):
    project_params = dict(type='page', label='global:{}'.format(global_label))
    return requests.get(_get_confluence_uri('prototype/1/search/site'),
                        params=project_params,
                        headers=get_headers(),
                        auth=get_auth()).json()


def _extract_id_set(response):
    return set([item['id'] for item in response['result']])


def _intersect_and_extract_single_id(response1, response2):
    intersection = _extract_id_set(response1).intersection(_extract_id_set(response2))
    if len(intersection) != 1:
        raise Exception('Should have found exactly one id, instead found {}'.format(intersection))
    return intersection.pop()


def _intersect_and_extract_single_id_or_none(response1, response2):
    intersection = _extract_id_set(response1).intersection(_extract_id_set(response2))
    if len(intersection) > 1:
        raise Exception('Should have found at most one id, instead found {}'.format(intersection))
    return (list(intersection) + [None])[0]


def get_project_response(project_name):
    return _get_confluence_global_response(project_name.lower())


def get_release_notes_page_id(project_name):
    release_notes_response = _get_confluence_global_response('release-notes')
    project_response = get_project_response(project_name)
    return _intersect_and_extract_single_id(release_notes_response, project_response)


def get_release_notes_header_page_id(project_name):
    release_notes_header_response = _get_confluence_global_response('release-notes-header')
    project_response = get_project_response(project_name)
    return _intersect_and_extract_single_id_or_none(release_notes_header_response, project_response)


def get_release_notes_footer_page_id(project_name):
    release_notes_footer_response = _get_confluence_global_response('release-notes-footer')
    project_response = get_project_response(project_name)
    return _intersect_and_extract_single_id_or_none(release_notes_footer_response, project_response)


def get_page_contents(page_id):
    page = requests.get(_get_confluence_uri('api/content/{}?expand=body.export_view,version.number'.format(page_id)),
                        auth=get_auth(),
                        headers=get_headers()).json()
    return page['body']['export_view']['value'].replace(u'\xc3\x82', '').replace(u'\xc2\xa0', '')


def get_page_storage(page_id):
    page = requests.get(_get_confluence_uri('api/content/{}?expand=body.storage,version.number'.format(page_id)),
                        auth=get_auth(),
                        headers=get_headers()).json()
    return page['body']['storage']['value'].replace(u'\xc3\x82', '').replace(u'\xc2\xa0', '')


def update_page_contents(page_id, body):
    page = requests.get(_get_confluence_uri('api/content/{}?expand=body.view,version.number,ancestors'.format(page_id)),
                        auth=get_auth()).json()
    data = dict(version=dict(number=page['version']['number']+1),
                id=page['id'], title=page['title'], type='page',
                body=dict(storage=dict(representation='storage', value=body)))
    if page['ancestors']:
        data['ancestors'] = [dict(id=page['ancestors'][-1]['id'])]
    requests.put(_get_confluence_uri('api/content/{}'.format(page_id)), json=data, auth=get_auth()).raise_for_status()


def iter_attachments(page_id, start=0, limit=50):
    data = requests.get(_get_confluence_uri('api/content/{page_id}/child/attachment'.format(page_id=page_id)),
                        headers=get_headers(),
                        auth=get_auth(),
                        params={'expand': 'id', 'start':start, 'limit':limit}).json()
    for item in data['results']:
        yield dict(title=item['title'], link=item['_links']['download'].split('?')[0] + '?api=v2')
    if len(data['results']) == limit:
        for _tuple in iter_attachments_tuple(page_id, start=start+limit, limit=limit):
            yield _tuple



import requests
import logging

logger = logging.getLogger(__name__)


class Plugin(object):
    def __init__(self, plugin_data):
        self._plugin_data = plugin_data
        self._name = self._plugin_data['name']
        self._key = self._plugin_data['key']
        self._installed_version = self._plugin_data['version']

    def get_info(self):
        return get("{}".format(self._plugin_data['links']['self']))

    def get(self, uri):
        return get("{}/{}".format(self._plugin_data['links']['self'], uri))

    def get_license_data(self):
        try:
            return self.get("license")
        except requests.HTTPError, error:
            logger.warn(error)
        return dict()

    def get_summary(self):
        return self.get("summary")

    @property
    def key(self):
        return self._key

    @property
    def installed_version(self):
        return self._installed_version

    @property
    def name(self):
        return self._name

    def get_info_on_marketplace(self):
        try:
            return get(self._plugin_data['links']['self'].replace("plugins/1.0/", "plugins/1.0/available/"))
        except requests.HTTPError, error:
            logger.warn(error)
        return dict()

    @property
    def marketplace_version(self):
        return self.get_info_on_marketplace().get('version')

    def __repr__(self):
        return "<Plugin: {}, v{}>".format(self._name, self._installed_version)

    @property
    def type(self):
        return 'User' if self._plugin_data['userInstalled'] else 'System'

    def is_enabled(self):
        return self._plugin_data['enabled']

    def is_update_available(self):
        return self.marketplace_version and self.marketplace_version != self.installed_version

    @property
    def expires_in(self):
        from datetime import datetime
        if not self.license_details.end_date:
            return False
        return (self.license_details.end_date - datetime.now()).days

    def is_license_expires_soon(self, days=60):
        if not self.expires_in:
            return False
        return self.expires_in <= days

    @property
    def license_details(self):
        from dateutil.parser import parse
        from munch import Munch
        license_data = self.get_license_data()
        return Munch(type=license_data.get('licenseType', '').lower(),
                     evaluation=license_data.get('evaluation'),
                     start_date=parse(license_data['creationDateString']) if license_data.get('creationDateString') else None,
                     end_date=parse(license_data['maintenanceExpiryDateString']) if license_data.get('maintenanceExpiryDateString') else None,
                     contract_number=license_data.get('supportEntitlementNumber'),
                     valid=license_data.get('valid'))


def get(uri, *args, **kwargs):
    from infi.jira_cli.jira_adapter import get_jira
    jira = get_jira()
    respnose = requests.get("{}{}".format(jira._options['server'], uri), auth=requests.auth.HTTPBasicAuth(*jira._session.auth))
    respnose.raise_for_status()
    return respnose.json()


def get_plugins():
    return [Plugin(item) for item in get("/rest/plugins/1.0/")['plugins']]

def get_available_upgrades():
    return get("/rest/plugins/1.0/available/upgrades")
